/***************************************************************************
    MboxHandler - Mbox-handling routines

    author:  Ulrich Krebs
    email :  ukrebs@freenet.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

import java.io.*;
import java.util.*;
import java.text.*;

class MboxHandler
{
	private Vector senders;
	private Vector receivers;
	private Vector subjects;
	private Vector dates;
	private Vector contentTypes;
	private Vector contentEncodings;
	private Vector contentStarts;
	private Vector contentEnds;
	private Vector contentFileNames;
	private Vector msgOffsets;
	private Vector sorting;
	private File file;
	private int numMessages=0;
	private LineReader br=null;

	public MboxHandler(String directory, String fileName)
	{
		this(new File(directory+fileName));
	}
	
	public MboxHandler(File f)
	{
		String line,lastLine="";
		boolean isHeader=false;

		file=f;
		senders=new Vector();
		receivers=new Vector();
		subjects=new Vector();
		dates=new Vector();
		contentTypes=new Vector();
		contentEncodings=new Vector();
		contentStarts=new Vector();
		contentEnds=new Vector();
		contentFileNames=new Vector();
		msgOffsets=new Vector();
		
		try
		{
			br=new LineReader(file);
			long overallTime=0;
			long startTime=System.currentTimeMillis();
			while((line=br.readLine())!=null)
			{
				if(line.regionMatches(false,0,"From ",0,5))
				{
					int curLine=br.getLineNr()-1;
					msgOffsets.addElement(new Integer(curLine));
					numMessages++;
		
					MailHandler mh=new MailHandler(br,curLine+1);
					
					if(mh.getSender()!=null)
					{
						senders.addElement(mh.getSender());
					}
					else
					{
						senders.addElement(new String("<unknown>"));
					}
					
					if(mh.getReceiver()!=null)
					{
						receivers.addElement(mh.getReceiver());
					}
					else
					{
						receivers.addElement(new String("<unknown>"));
					}
					
					if(mh.getSubject()!=null)
					{
						subjects.addElement(mh.getSubject());
					}
					else
					{
						subjects.addElement(new String("<unknown>"));
					}
					
					contentTypes.addElement(mh.getContentTypes());
					contentEncodings.addElement(mh.getContentEncodings());
					contentStarts.addElement(mh.getContentStarts());
					contentEnds.addElement(mh.getContentEnds());
					contentFileNames.addElement(mh.getContentFileNames());
					
					if(mh.getSendDate()!=null)
					{
						dates.addElement(mh.getSendDate());
					}
					else if(mh.getReceiveDate()!=null)
					{
						dates.addElement(mh.getReceiveDate());
					}
					else
					{
						dates.addElement(new Date(0));
					}
				}
				lastLine=line;
			}

			addFileNames();
			
			overallTime+=System.currentTimeMillis()-startTime;
			System.out.println(overallTime);

			sorting=new Vector();
			for(int i=0;i<numMessages;i++) sorting.addElement(new Integer(i));
			sorting=indexSort(sorting,dates);
		}
		catch(FileNotFoundException e)
		{
		}
		catch(IOException e)
		{
		}
	}
	
	public int getNumMessages()
	{
		return numMessages;
	}

	public String getSender(int index)
	{
		return (String)senders.elementAt(((Integer)sorting.elementAt(index)).intValue());
	}

	public String getReceiver(int index)
	{
		return (String)receivers.elementAt(((Integer)sorting.elementAt(index)).intValue());
	}

	public String getFormattedDate(int index)
	{
		Date date=(Date)dates.elementAt(((Integer)sorting.elementAt(index)).intValue());
		String formattedDate;
		if(date.getTime() != 0L)
		{
			SimpleDateFormat formatter=new SimpleDateFormat("dd MMMM yyyy, HH:mm:ss Z");
			formattedDate=formatter.format(date);
		}
		else
		{
			formattedDate="<unknown>";
		}
		return formattedDate;
	}

	public Date getDate(int index)
	{
		return (Date)dates.elementAt(((Integer)sorting.elementAt(index)).intValue());
	}
	
	public String getSubject(int index)
	{
		return (String)subjects.elementAt(((Integer)sorting.elementAt(index)).intValue());
	}
	
	public String getContentType(int index)
	{
		return (String)contentTypes.elementAt(((Integer)sorting.elementAt(index)).intValue());
	}

	public String getMessageContent(int index)
	{
		StringBuffer outBuf=new StringBuffer("");
		String fileName;
		String[] knownContents={"TEXT", "TEXT/PLAIN", "TEXT/HTML"};
		
		index=((Integer)sorting.elementAt(index)).intValue();

		int numParts=((int[])contentStarts.elementAt(index)).length;

		int k=0;
		do
		{
			for(int i=0;i<numParts;i++)
			{
				String contentType=((String[])contentTypes.elementAt(index))[i];
				if(contentType==null)
				{
					contentType="TEXT/PLAIN";
				}
				else contentType=contentType.toUpperCase();
				
				
				if(contentType.equals(knownContents[k]))
				{
					if(outBuf.length()>0)
					{
						outBuf.append("\r\n");
						outBuf.append("--------- Next Part");
						if((fileName=((String[])contentFileNames.elementAt(index))[i])!=null)
						{
							outBuf.append(": \""+fileName+"\"");
						}
						outBuf.append(" ---------");
						outBuf.append("\r\n\r\n");
					}
	
					String contentEncoding=((String[])contentEncodings.elementAt(index))[i];
					if(contentEncoding==null)
					{
						contentEncoding="7BIT";
					}
			
					int contentStart=((int[])contentStarts.elementAt(index))[i];
					int contentEnd=((int[])contentEnds.elementAt(index))[i];
			
					try
					{
						br.setLineNr(contentStart);
							
						String line;
							
						for(int j=contentStart;j<=contentEnd;j++)
						{
							line=br.readLine();
							outBuf.append(line);
							if(!contentEncoding.equalsIgnoreCase("BASE64"))
							{
								if(j<contentEnd) outBuf.append("\r\n");
							}
						}
						
						if(contentEncoding.equalsIgnoreCase("QUOTED-PRINTABLE"))
						{
							outBuf=new StringBuffer(QuotedPrintable.decode(outBuf.toString()));
						}
						if(contentEncoding.equalsIgnoreCase("BASE64"))
						{
							outBuf=new StringBuffer(new String(Base64.decode(outBuf.toString().getBytes())));
						}
					}
					catch(FileNotFoundException e)
					{
					}
					catch(IOException e)
					{
					}
				}
			}
			k++;
		}
		while(outBuf.length()==0 && k<knownContents.length);
		
		if(knownContents[k-1].equals("TEXT/HTML"))
		{
			HTMLFileHandler fh=new HTMLFileHandler(outBuf.toString());
			outBuf=new StringBuffer(fh.getPlainText());
		}
		return outBuf.toString();
	}
	
	private void addFileNames()
	{
		for(int i=0;i<numMessages;i++)
		{
			int numContents=((int[])contentStarts.elementAt(i)).length;
			String[] fileNames=(String[])contentFileNames.elementAt(i);
			String[] types=(String[])contentTypes.elementAt(i);

			for(int j=0;j<numContents;j++)
			{
				if(fileNames[j]==null)
				{
					if((types[j]!=null && types[j].equalsIgnoreCase("MESSAGE/RFC822")))
					{
						int contentStart=((int[])contentStarts.elementAt(i))[j];
						int contentEnd=((int[])contentEnds.elementAt(i))[j];
						MailHandler mh=new MailHandler(br,contentStart,contentEnd);
						String attachmentSubject;
						if(mh.getSubject()!=null)
						{
							attachmentSubject=mh.getSubject();
						}
						else
						{
							attachmentSubject="";
						}
						((String[])contentFileNames.elementAt(i))[j]=makeFilename("",attachmentSubject)+".eml";
					}
				}
			}
		}
	}
	
	public String[] getAttachmentFileNames(int index)
	{
		index=((Integer)sorting.elementAt(index)).intValue();
		
		String[] fileNames=(String[])contentFileNames.elementAt(index);
		String[] types=(String[])contentTypes.elementAt(index);
		int numFileNames=0;
		
		for(int i=0;i<fileNames.length;i++)
		{
			if(fileNames[i]!=null)
			{
				numFileNames++;
			}
		}
		
		String[] out=new String[numFileNames];
		int j=0;
		
		for(int i=0;i<fileNames.length;i++)
		{
			if(fileNames[i]!=null)
			{
				out[j++]=fileNames[i];
			}
		}
		
		return out;
	}
	
	public String getFilename()
	{
		return(file.getName());
	}

	public String getDirectory()
	{
		return(new File(file.getAbsolutePath()).getParent());
	}

	public String getPath()
	{
		return(file.getAbsolutePath());
	}
	
	public File getFile()
	{
		return(file);
	}
	
	private String makeFilename(String name, String subject)
	{
		int index;
		
		if(name==null) name=new String("");
		if(subject==null) subject=new String("");
		
		name=name.trim();
		subject=subject.trim();
		
		name=name.replaceAll("[\\x01-\\x1F]|[<>|?/\\\\\":*\\r\\n]"," ");
		subject=subject.replaceAll("[\\x01-\\x1F]|[<>|?/\\\\\":*\\r\\n]"," ");
		name=name.replaceAll(" {2,}"," ");
		subject=subject.replaceAll(" {2,}"," ");

		if((index=name.indexOf("@"))>=0)
		{
			name=name.substring(0,index);
		}
		
		if(subject.equals("")) subject="No Subject";

		if(name.length()+subject.length()>60)
		{
			if(subject.length()>40)
			{
				subject=subject.substring(0,40);
			}
			if(name.length()+subject.length()>60)
			{
				name=name.substring(0,20);
			}
		}
		
		if(name.length()==0) return(subject);
		
		return(name+" - "+subject);
	}
	
	public int saveAsEml(int[] messages, boolean useReceiver, String directory)
	{
		Vector toSave=new Vector();
		Vector toSaveOffsets=new Vector();
		int successfulExported=0;
		
		for(int i=0;i<messages.length;i++)
		{
			Integer index=(Integer)sorting.elementAt(messages[i]);
			toSave.addElement(index);
		}

		File outputFile;

		try
		{
			for(int i=0;i<toSave.size();i++)
			{
				int index=((Integer)toSave.elementAt(i)).intValue();
				int msgOffset=((Integer)msgOffsets.elementAt(index)).intValue();
				Date msgDate=(Date)dates.elementAt(index);

				int msgOffsetNext;
				if(index<msgOffsets.size()-1)
				{
					msgOffsetNext=((Integer)msgOffsets.elementAt(index+1)).intValue();
				}
				else
				{
					msgOffsetNext=-1;
				}
				
				br.setLineNr(msgOffset+1);
				boolean msgEnd=false;
				
				String baseFileName;
				if(useReceiver)
				{
					baseFileName=makeFilename((String)receivers.elementAt(index),(String)subjects.elementAt(index));
				}
				else
				{
					baseFileName=makeFilename((String)senders.elementAt(index),(String)subjects.elementAt(index));
				}
				String fileName=baseFileName+".eml";
				
				File outFile=new File(directory);
				if(!outFile.exists())
				{
					outFile.mkdirs();
				}
				int j=1;
				while((outFile=new File(directory,fileName)).exists())
				{
					fileName=baseFileName+"("+j+")"+".eml";
					j++;
				}
				
				String line;
				
				try
				{
					BufferedWriter bw=new BufferedWriter(new FileWriter(outFile));
					while(!msgEnd && (line=br.readLine())!=null)
					{
						bw.write(line,0,line.length());
						bw.newLine();
						if(msgOffsetNext>=0 && br.getLineNr()>=msgOffsetNext-1)
						{
							msgEnd=true;
						}
					}
					bw.close();
					if(msgDate!=null && msgDate.getTime()>=0)
					{
						outFile.setLastModified(msgDate.getTime());
					}
					successfulExported++;
				}
				catch(IOException e)
				{
					System.out.println(e);
				}
			}
		}
		catch(FileNotFoundException e)
		{
		}
		catch(IOException e)
		{
		}
		return successfulExported;
	}

	public int saveAttachments(int index, int[] attachments, String directory)
	{
		int successfulSaved=0;
		
		index=((Integer)sorting.elementAt(index)).intValue();
		Date msgDate=(Date)dates.elementAt(index);
		String[] fileNames=(String[])contentFileNames.elementAt(index);

		try
		{
			for(int i=0;i<attachments.length;i++)
			{
				int j=0,k=0;
				for(j=0;j<fileNames.length;j++)
				{
					if(fileNames[j]!=null)
					{
						if(k==attachments[i]) break;
						k++;
					}
				}
				if(k!=attachments[i]) continue;

				int contentStart=((int[])contentStarts.elementAt(index))[j];
				int contentEnd=((int[])contentEnds.elementAt(index))[j];
				String contentEncoding=((String[])contentEncodings.elementAt(index))[j];
				if(contentEncoding==null)
				{
					contentEncoding="7BIT";
				}

				String fileName=fileNames[j];
				String baseFileName=fileName;
				
				File outFile=new File(directory);
				if(!outFile.exists())
				{
					outFile.mkdirs();
				}
				
				j=1;
				while((outFile=new File(directory,fileName)).exists())
				{
					fileName="("+j+")"+baseFileName;
					j++;
				}
				
				String line;
				int curLine;
				br.setLineNr(contentStart);
				
				try
				{
					StringBuffer out=new StringBuffer();
					if(contentEncoding.equalsIgnoreCase("BASE64"))
					{
						while((curLine=br.getLineNr())<=contentEnd && (line=br.readLine())!=null)
						{
							out.append(line);
						}
						byte[] outBytes=out.toString().getBytes();
						outBytes=Base64.decode(outBytes);
						FileOutputStream fo=new FileOutputStream(outFile);
						fo.write(outBytes);
						fo.close();
					}
					else if(contentEncoding.equalsIgnoreCase("QUOTED-PRINTABLE"))
					{
						while((curLine=br.getLineNr())<=contentEnd && (line=br.readLine())!=null)
						{
							out.append(line);
							if(curLine<contentEnd) out.append("\r\n");
						}
						String outString=QuotedPrintable.decode(out.toString());
						BufferedWriter bw=new BufferedWriter(new FileWriter(outFile));
						bw.write(outString);
						bw.close();
					}
					else if(contentEncoding.equalsIgnoreCase("X-UUENCODE") || contentEncoding.equalsIgnoreCase("UUENCODE"))
					{
						while((curLine=br.getLineNr())<=contentEnd && (line=br.readLine())!=null)
						{
							out.append(line);
							if(curLine<contentEnd) out.append("\r\n");
						}
						byte[] outBytes=Uuencode.decode(out.toString());
						FileOutputStream fo=new FileOutputStream(outFile);
						fo.write(outBytes);
						fo.close();
					}
					else
					{
						while((curLine=br.getLineNr())<=contentEnd && (line=br.readLine())!=null)
						{
							out.append(line);
							if(curLine<contentEnd) out.append("\r\n");
						}
						BufferedWriter bw=new BufferedWriter(new FileWriter(outFile));
						bw.write(out.toString());
						bw.close();
					}
					if(msgDate!=null)
					{
						outFile.setLastModified(msgDate.getTime());
					}
					successfulSaved++;
				}
				catch(IOException e)
				{
				}
			}
		}
		catch(FileNotFoundException e)
		{
		}
		catch(IOException e)
		{
		}
		return successfulSaved;
	}
	
	public void close()
	{
		try
		{
			br.close();
		}
		catch(IOException e)
		{
		}
	}

	public void sort(String criterion)
	{
		Vector toSort;
		if(criterion.equals("Date"))
		{
			toSort=dates;
		}
		else if(criterion.equals("Sender"))
		{
			toSort=senders;
		}
		else if(criterion.equals("Receiver"))
		{
			toSort=receivers;
		}
		else if(criterion.equals("Subject"))
		{
			toSort=subjects;
		}
		else return;
		
		sorting=indexSort(sorting,toSort);
	}

	private Vector indexSort(Vector index, Vector items)
	{
		Vector buffer=new Vector();
		Vector newIndex=new Vector();
		
		boolean isString=false;
		
		if(index.size()>0)
		{
			isString=items.elementAt(0) instanceof String;
		}
		
		for(int i=0;i<index.size();i++)
		{
			Object item;
			if(isString)
			{
				item=((String)items.elementAt(((Integer)index.elementAt(i)).intValue())).toUpperCase();
			}
			else
			{
				item=items.elementAt(((Integer)index.elementAt(i)).intValue());
			}
			int insertionPoint=java.util.Collections.binarySearch((List)buffer,item);
			if(insertionPoint<0) insertionPoint=-insertionPoint-1;
			buffer.insertElementAt(item,insertionPoint);
			newIndex.insertElementAt(index.elementAt(i),insertionPoint);
		}
		return newIndex;
	}
}
